__all__ = ['SlicePicker', 'Collage', 'Carpet']

import traceback
import warnings
from collections import Iterable

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.image import AxesImage
from mrivis import config as cfg
from mrivis.utils import check_num_slices, check_views, load_image_from_disk, \
    row_wise_rescale, save_figure, check_matching_dims


class SlicePicker(object):
    """
    Class to pick non-empty slices along the various dimensions for a given image.

    The term `slice` here refers to one cross-section in a 3D image,
        towards which this class is designed for.
        However there are no explicit restrictions placed on dicing N=4+ array
        and receiving a n-1 dim array.
    """

    def __init__(self,
                 image_in,
                 view_set=cfg.view_set_default,
                 num_slices=cfg.num_slices_default,
                 sampler=cfg.sampler_default,
                 min_density=cfg.min_density_default):
        """
        Class to pick non-empty slices along the various dimensions for a given image.

        Parameters
        ----------
        image_in : ndarray
            3D array to be sliced.
            there are no explicit restrictions placed on number of dimensions for image_in,
             to get a n-1 dim array, but appropriate reshaping may need to be performed.

        view_set : iterable
            List of integers selecting the dimesnions to be sliced.

        num_slices : int or iterable of size as view_set
            Number of slices to be selected in each view.

        sampler : str or list or callable
            selection strategy: to identify the type of sampling done to select the slices to return.
            All sampling is done between the first and last non-empty slice in that view/dimension.

            - if 'linear' : linearly spaced slices
            - if list, it is treated as set of percentages at which slices to be sampled
                (must be in the range of [1-100], not [0-1]).
                This could be used to more/all slices in the middle e.g. range(40, 60, 5)
                    or at the end e.g. [ 5, 10, 15, 85, 90, 95]
            - if callable, it must take a 2D image of arbitray size, return True/False
                to indicate whether to select that slice or not.
                Only non-empty slices (atleas one non-zero voxel) are provided as input.
                Simple examples for callable could be based on
                1) percentage of non-zero voxels > x etc
                2) presence of desired texture ?
                3) certain properties of distribution (skewe: dark/bright, energy etc) etc

                If the sampler returns more than requested `num_slices`,
                    only the first num_slices will be selected.

        min_density : float or None
            mininum density of non-zero voxels within a given slice to consider it non-empty
            Default: 0.01 (1%).
            if None, include all slices.

        """

        if len(image_in.shape) < 3:
            raise ValueError('Image must be atleast 3D')
        else:
            self._image = image_in

        self._check_min_density(min_density)

        self._image_shape = np.array(self._image.shape)
        self.view_set = check_views(view_set, max_views=len(self._image_shape))
        self.num_slices = check_num_slices(num_slices,
                                           img_shape=self._image_shape[self.view_set],
                                           num_dims=len(self.view_set))

        self._verify_sampler(sampler)

        self._pick_slices()  # creates self._slices

    def _verify_sampler(self, sampler):

        if isinstance(sampler, str):
            sampler = sampler.lower()
            if sampler not in ['linear', ]:
                raise ValueError('Sampling strategy: {} not implemented.'.format(sampler))
            self._sampler = sampler
            self._sampling_method = 'linear'
        elif isinstance(sampler, Iterable):
            if any([index < 0 or index > 100 for index in sampler]):
                raise ValueError('sampling percentages must be in  [0-100]% range')
            if len(sampler) > min(self.num_slices):
                self.num_slices = np.maximum(self.num_slices, len(sampler))
            self._sampler = np.array(sampler)
            self._sampling_method = 'percentage'
        elif callable(sampler):
            # checking if the callable returns a bool
            for view in self.view_set:
                middle_slice = int(self._image_shape[view] / 2)
                if not isinstance(sampler(self._get_axis(self._image, view, middle_slice)), bool):
                    raise ValueError('sampler callable must return a boolean value (True/False)')

            self._sampler = sampler
            self._sampling_method = 'callable'
        else:
            raise NotImplementedError('Invalid choice for sampler! Choose one of: '
                                      'linear, percentage or callable')

    def _check_min_density(self, min_density):
        """Validator to ensure proper usage."""

        if min_density is None:
            self._min_density = -np.Inf
        elif (isinstance(min_density, float) and (0.0 <= min_density < 1.0)):
            self._min_density = min_density
        else:
            raise ValueError('min_density must be float and be >=0.0 and < 1.0')

    def _pick_slices(self):
        """
        Picks the slices to display in each dimension/view,
            skipping any empty slices (without any data at all).
        """

        self._slices = list()
        self._slices_by_dim = list()
        for view, ns_in_view in zip(self.view_set, self.num_slices):
            # discarding completely empty or almost empty slices.
            dim_size = self._image_shape[view]
            non_empty_slices = np.array([sl for sl in range(dim_size) if self._not_empty(view, sl)])

            # sampling according to the chosen strategy
            slices_dim = self._sample_slices_in_dim(view, ns_in_view, non_empty_slices)
            #  the following loop is needed to preserve order, while eliminating duplicates
            # list comprehension over a set(slices_dim) wouldn't preserve order
            uniq_slices = list()
            for sn in slices_dim:
                if (0 <= sn < dim_size) and sn not in uniq_slices:
                    uniq_slices.append(sn)

            self._slices_by_dim.append(slices_dim)
            # adding view and slice # at the same time..
            self._slices.extend([(view, sn) for sn in slices_dim])

    def _not_empty(self, view, slice_):
        """Checks if the density is too low. """

        img2d = self._get_axis(self._image, view, slice_)
        return (np.count_nonzero(img2d) / img2d.size) > self._min_density

    def _sample_slices_in_dim(self, view, num_slices, non_empty_slices):
        """Samples the slices in the given dimension according the chosen strategy."""

        if self._sampling_method == 'linear':
            return self._linear_selection(non_empty_slices=non_empty_slices, num_slices=num_slices)
        elif self._sampling_method == 'percentage':
            return self._percent_selection(non_empty_slices=non_empty_slices)
        elif self._sampling_method == 'callable':
            return self._selection_by_callable(view=view, non_empty_slices=non_empty_slices,
                                               num_slices=num_slices)
        else:
            raise NotImplementedError('Invalid state for the class!')

    def _linear_selection(self, non_empty_slices, num_slices):
        """Selects linearly spaced slices in given"""

        num_non_empty = len(non_empty_slices)

        # # trying to skip 5% slices at the tails (bottom clipping at 0)
        # skip_count = max(0, np.around(num_non_empty * 0.05).astype('int16'))
        # # only when possible
        # if skip_count > 0 and (num_non_empty - 2 * skip_count > num_slices):
        #     non_empty_slices = non_empty_slices[skip_count: -skip_count]
        #     num_non_empty = len(non_empty_slices)

        sampled_indices = np.linspace(0, num_non_empty, num=min(num_non_empty, num_slices),
                                      endpoint=False)
        slices_in_dim = non_empty_slices[np.around(sampled_indices).astype('int64')]

        return slices_in_dim

    def _percent_selection(self, non_empty_slices):
        """Chooses slices at a given percentage between the first and last non-empty slice."""

        return np.around(self._sampler * len(non_empty_slices) / 100).astype('int64')

    def _selection_by_callable(self, view, num_slices, non_empty_slices):
        """Returns all the slices selected by the given callable."""

        selected = [sl for sl in non_empty_slices
                    if self._sampler(self._get_axis(self._image, view, sl))]

        return selected[:num_slices]

    def _get_axis(self, array, axis, slice_num,
                  extended=False,
                  transpose=True):
        """Returns a fixed axis"""

        slice_list = [slice(None)] * array.ndim
        slice_list[axis] = slice_num
        slice_data = array[tuple(slice_list)]
        if transpose:  # for proper orientation
            slice_data = slice_data.T

        if not extended:
            # return just the slice data
            return slice_data
        else:
            # additionally include which dim and which slice num
            return axis, slice_num, slice_data

    def get_slice_indices(self):
        """Returns indices for the slices selected (each a tuple : (dim, slice_num))"""

        return self._slices

    def get_slices(self, extended=False):
        """Generator over all the slices selected, each time returning a cross-section.

        Parameters
        ----------

        extended : bool
            Flag to return just slice data (default, extended=False), or
                return a tuple of axis, slice_num, slice_data (extended=True)

        Returns
        -------

        slice_data : an image (just slice data, default, with extended=False), or
                a tuple of axis, slice_num, slice_data (extended=True)

        """

        for dim, slice_num in self._slices:
            yield self._get_axis(self._image, dim, slice_num, extended=extended)

    def get_slices_multi(self, image_list, extended=False):
        """Returns the same cross-section from the multiple images supplied.

        All images must be of the same shape as the original image defining this object.

        Parameters
        ----------

        image_list : Iterable
            containing atleast 2 images

        extended : bool
            Flag to return just slice data (default, extended=False), or
                return a tuple of axis, slice_num, slice_data (extended=True)

        Returns
        -------

        tuple_slice_data : tuple of one slice from each image in the input image list
                            Let's denote it by as TSL.
                        if extended=True, returns tuple(axis, slice_num, TSL)

        """

        # ensure all the images have the same shape
        for img in image_list:
            if not check_matching_dims(img, self._image):
                raise ValueError('Supplied images do match in size.'
                                 ' This image has dimensions: {}'
                                 ' They must all have: {}'
                                 ''.format(img.shape, self._image_shape))

        for dim, slice_num in self._slices:
            multiple_slices = (self._get_axis(img, dim, slice_num) for img in image_list)
            if not extended:
                # return just the slice data
                yield multiple_slices
            else:
                # additionally include which dim and which slice num
                # not using extended option in get_axis, to avoid complicating unpacking
                yield dim, slice_num, multiple_slices

    def save_as_gif(self, gif_path, duration=0.25):
        """Package the selected slices into a single GIF for easy sharing and display (on web etc).

        You must install imageio module separately to use this feature.

        Parameters
        ----------
        gif_path : str
            Output path for the GIF image

        duration : float
            Duration of display of each frame in GIF, in sec.

        """

        import imageio
        gif_data = [img for img in self.get_slices()]
        # using img.astype(np.uint32) is leaving noticable artefacts,
        #   depending on imageio inner conversion, which is rescaling okay

        # TODO deal with differing sizes of slices, padding with zeros or white??

        with warnings.catch_warnings():
            # ignoring the warning for conversion to uint8
            warnings.simplefilter('ignore')
            imageio.mimsave(gif_path, gif_data, duration=duration)

    def __iter__(self):
        """Returns the next panel, and the associated dimension and slice number"""

        return iter(self._slices)

    def __len__(self):
        """Returns the total number of slices across all the views."""

        return len(self._slices)

    def __format__(self, format_spec='s'):
        """various formats"""

        if format_spec in ['s', 'simple']:
            return self.__str__()
        elif format_spec in ['f', 'full']:
            return self.__repr__()
        else:
            return 'invalid format requsted!!'

    def __str__(self):

        return 'views : {}\n' \
               '#slices: {}\n' \
               'sampler: {}'.format(self.view_set, self.num_slices, self._sampling_method)

    def __repr__(self):

        dim_repr = list()
        for ix, vw in enumerate(self.view_set):
            dim_repr.append('{} slices in dim {} : '
                            '{}'.format(len(self._slices_by_dim[ix]),
                                        vw, self._slices_by_dim[ix]))
        return '\n'.join(dim_repr)


class MiddleSlicePicker(SlicePicker):
    """Convenience class to select the classic one middle slice from all views."""

    def __init__(self, image):
        """Returns the middle slice from all views in the image.

        Parameters
        -----------

        attach_image : ndarray
            The image to be attached to the collage, once it is created.
            Must be atleast 3d.

        """

        super().__init__(image_in=image,
                         view_set=cfg.view_set_default,
                         num_slices=1,
                         sampler=(50,), )


class Collage(object):
    """
    Class exhibiting multiple slices from a 3D image,
        with convenience routines handling all the cross-sections as a single set.
    """

    def __init__(self,
                 view_set=cfg.view_set_default,
                 num_rows=cfg.num_rows_per_view_default,
                 num_slices=cfg.num_slices_default,
                 sampler=cfg.sampler_default,
                 attach_image=None,
                 bounding_rect=cfg.bounding_rect_default,
                 fig=None,
                 figsize=cfg.figsize_default,
                 display_params=None,
                 ):
        """
        Class exhibiting multiple slices from a 3D image,
        with convenience routines handling all the cross-sections as a single set.

        Once created with certain `display_params` (containing vmin and vmax),
            this class does NOT automatically rescale the data, as you attach different images.
            Ensure the input images are rescaled to [0, 1] BEFORE attaching.

        Parameters
        ----------

        view_set : iterable
            List of integers selecting the dimesnions to be sliced.

        num_slices : int or iterable of size as view_set
            Number of slices to be selected in each view.

        num_rows : int
            Number of rows per view.

        sampler : str or list or callable
            selection strategy to identify the type of sampling done
            to select the slices to return. All sampling is done between
            the first and last non-empty slice in that view/dimension.

            - if 'linear' : linearly spaced slices
            - if list, it is treated as set of percentages at which slices to be sampled
                (must be in the range of [1-100], not [0-1]).
                This could be used to more/all slices in the middle e.g. range(40, 60, 5)
                    or at the end e.g. [ 5, 10, 15, 85, 90, 95]
            - if callable, it must take a 2D image of arbitray size, return True/False
                to indicate whether to select that slice or not.
                Only non-empty slices (atleas one non-zero voxel) are provided as input.
                Simple examples for callable could be based on
                1) percentage of non-zero voxels > x etc
                2) presence of desired texture ?
                3) certain properties of distribution (skewe: dark/bright, energy etc) etc

                If the sampler returns more than requested `num_slices`,
                    only the first num_slices will be selected.

        attach_image : ndarray
            The image to be attached to the collage, once it is created.
            Must be atleast 3d.

        display_params : dict
            dict of keyword parameters that can be passed to matplotlib's `Axes.imshow()`

        fig : matplotlib.Figure
            figure handle to create the collage in.
            If not specified, creates a new figure.

        figsize : tuple of 2
            Figure size (width, height) in inches.

        bounding_rect : tuple of 4
            The rectangular area to bind the collage to (in normalized figure coordinates)

        """

        self.view_set = check_views(view_set, max_views=3)
        self.num_slices = check_num_slices(num_slices, img_shape=None,
                                           num_dims=len(self.view_set))
        # TODO find a way to validate the input-- using utits.verify_sampler commonly?
        self.sampler = sampler

        if display_params is None:
            self.display_params = dict(interpolation='none', origin='lower',
                                       aspect='equal', cmap='gray', vmin=0.0, vmax=1.0)
        else:
            self.display_params = display_params
        self._make_layout(fig, figsize, num_rows, bounding_rect=bounding_rect)
        if attach_image is not None:
            self.attach(attach_image)
        else:
            self._data_attached = False

    def _make_layout(self,
                     fig,
                     figsize=cfg.figsize_default,
                     num_rows_per_view=cfg.num_rows_per_view_default,
                     bounding_rect=cfg.bounding_rect_default,
                     grid_pad=cfg.grid_pad_default,
                     axis_pad=cfg.axis_pad_default,
                     **axis_kwargs):

        plt.style.use('dark_background')
        if fig is None:
            self.fig = plt.figure(figsize=figsize)
        else:
            self.fig = fig

        total_num_rows = len(self.view_set) * num_rows_per_view
        total_num_panels = sum(self.num_slices)
        num_cols_per_row = int(np.ceil(total_num_panels / total_num_rows))

        left, bottom, width, height = bounding_rect
        avail_height = height - bottom
        num_views = len(self.view_set)
        height_each_view = (avail_height - num_views * grid_pad) / num_views
        effective_height_each_view = height_each_view + grid_pad

        self.grids = list()
        for ix, view in enumerate(self.view_set):
            rect = (left, bottom + ix * effective_height_each_view,
                    width, height_each_view)
            ax_grid = self._make_grid_of_axes(bounding_rect=rect, axis_pad=axis_pad,
                                              num_rows=num_rows_per_view,
                                              num_cols=num_cols_per_row,
                                              **axis_kwargs)
            self.grids.append(ax_grid)

        # flattened for easy access
        self.flat_grid = [ax for gg in self.grids for ax in gg]
        # create self.images with one image in each axis
        self._create_imshow_objects()

    def _make_grid_of_axes(self,
                           bounding_rect=cfg.bounding_rect_default,
                           num_rows=cfg.num_rows_per_view_default,
                           num_cols=cfg.num_cols_grid_default,
                           axis_pad=cfg.axis_pad_default,
                           commn_annot=None,
                           **axis_kwargs):
        """Creates a grid of axes bounded within a given rectangle."""

        axes_in_grid = list()
        extents = self._compute_cell_extents_grid(bounding_rect=bounding_rect,
                                                  num_cols=num_cols,
                                                  num_rows=num_rows, axis_pad=axis_pad)
        for cell_ext in extents:
            ax_cell = self.fig.add_axes(cell_ext, frameon=False, visible=False,
                                        **axis_kwargs)
            if commn_annot is not None:
                ax_cell.set_title(commn_annot)
            ax_cell.set_axis_off()
            axes_in_grid.append(ax_cell)

        return axes_in_grid

    @staticmethod
    def _compute_cell_extents_grid(bounding_rect=(0.03, 0.03, 0.97, 0.97),
                                   num_rows=2, num_cols=6,
                                   axis_pad=0.01):
        """
        Produces array of num_rows*num_cols elements each containing the rectangular extents of
        the corresponding cell the grid, whose position is within bounding_rect.
        """

        left, bottom, width, height = bounding_rect
        height_padding = axis_pad * (num_rows + 1)
        width_padding = axis_pad * (num_cols + 1)
        cell_height = float((height - height_padding) / num_rows)
        cell_width = float((width - width_padding) / num_cols)

        cell_height_padded = cell_height + axis_pad
        cell_width_padded = cell_width + axis_pad

        extents = list()
        for row in range(num_rows - 1, -1, -1):
            for col in range(num_cols):
                extents.append((left + col * cell_width_padded,
                                bottom + row * cell_height_padded,
                                cell_width, cell_height))

        return extents

    def _create_imshow_objects(self):
        """Turns off all the x and y axes in each Axis"""

        # uniform values for initial image can cause weird behaviour with normalization
        #       as imshow.set_data() does not automatically update the normalization!!
        # using random data is a better choice
        random_image = np.random.rand(20, 20)
        self.images = [None] * len(self.flat_grid)
        for ix, ax in enumerate(self.flat_grid):
            self.images[ix] = ax.imshow(random_image, **self.display_params)

    def show(self, grid=None):
        """Makes the collage visible.

        Parameters
        ----------

        grid : int or None
            index (into original ``view_set``) to which grid/view needs to be hidden

        """

        self._set_visible(True, grid_index=grid)

    def attach(self,
               image_in,
               sampler=None,
               show=True):
        """Attaches the relevant cross-sections to each axis.

        Parameters
        ----------

        attach_image : ndarray
            The image to be attached to the collage, once it is created.
            Must be atleast 3d.

        sampler : str or list or callable
            selection strategy: to identify the type of sampling done to select the slices to return.
            All sampling is done between the first and last non-empty slice in that view/dimension.

            - if 'linear' : linearly spaced slices
            - if list, it is treated as set of percentages at which slices to be sampled
                (must be in the range of [1-100], not [0-1]).
                This could be used to more/all slices in the middle e.g. range(40, 60, 5)
                    or at the end e.g. [ 5, 10, 15, 85, 90, 95]
            - if callable, it must take a 2D image of arbitray size, return True/False
                to indicate whether to select that slice or not.
                Only non-empty slices (atleas one non-zero voxel) are provided as input.
                Simple examples for callable could be based on
                1) percentage of non-zero voxels > x etc
                2) presence of desired texture ?
                3) certain properties of distribution (skewe: dark/bright, energy etc) etc

                If the sampler returns more than requested `num_slices`,
                    only the first num_slices will be selected.

        show : bool
            Flag to request immediate display of collage

        """

        if len(image_in.shape) < 3:
            raise ValueError('Image must be atleast 3D')

        # allowing the choice of new sampling for different invocations.
        if sampler is None:
            temp_sampler = self.sampler
        else:
            temp_sampler = sampler

        slicer = SlicePicker(image_in=image_in,
                             view_set=self.view_set,
                             num_slices=self.num_slices,
                             sampler=temp_sampler)

        try:
            for img_obj, slice_data in zip(self.images, slicer.get_slices()):
                img_obj.set_data(slice_data)
        except:
            self._data_attached = False
            raise ValueError('unable to attach the given image data to current collage')
        else:
            self._data_attached = True

        # show all the axes
        if show:
            self.show()

    def transform_and_attach(self,
                             image_list,
                             func,
                             show=True):
        """
        Displays the transformed (combined) version of the cross-sections from each image,
            (same slice and dimension). So if you input n>=1 images, n slices are obtained
            from each image, which are passed to the func (callable) provided, and the
            result will be displayed in the corresponding cell of the collage.
            Useful applications:
            - input two images, a function to overlay edges of one image on the other
            - input two images, a function to mix them in a checkerboard pattern
            - input one image, a function to saturate the upper half of intensities
                (to increase contrast and reveal any subtle ghosting in slices)

        func must be able to receive as many arguments as many elements in image_list.
            if your func needs additional parameters, make them keyword arguments, and
            use functools.partial to obtain a new callable that takes in just the slices.

        Parameters
        -----------
        image_list : list or ndarray
            list of images or a single ndarray

        func : callable
            function to be applied on the input images (their slices)
                to produce a single slice to be displayed.

        show : bool
            flag to indicate whether make the collage visible.

        """

        if not callable(func):
            raise TypeError('func must be callable!')

        if not isinstance(image_list, (tuple, list)) and isinstance(image_list, np.ndarray):
            image_list = [image_list, ]

        if len(image_list) > 1:
            for ii in range(1, len(image_list)):
                if not check_matching_dims(image_list[ii], image_list[0]):
                    raise ValueError('All images must be of same shape!')
                if len(image_list[ii].shape) < 3:
                    raise ValueError('All images must be atleast 3D')

        slicer = SlicePicker(image_in=image_list[0],
                             view_set=self.view_set,
                             num_slices=self.num_slices)

        try:
            for img_obj, slice_list in zip(self.images,
                                           slicer.get_slices_multi(image_list)):
                img_obj.set_data(func(*slice_list))
        except:
            self._data_attached = False
            raise ValueError('unable to attach mix of given images to current collage')
        else:
            self._data_attached = True

        # show all the axes
        if show:
            self.show()

    def hide(self, grid=None):
        """Removes the collage from view.

        Parameters
        ----------

        grid : int or None
            index (into original ``view_set``) to which grid/view needs to be hidden

        """

        self._set_visible(False, grid_index=grid)

    def _set_visible(self, visibility, grid_index=None):
        """Sets the visibility property of all axes."""

        if grid_index is None:
            for ax in self.flat_grid:
                ax.set_visible(visibility)
        else:
            if grid_index < 0 or grid_index >= len(self.grids):
                raise IndexError('Valid indices : 0 to {}'.format(len(self.grids) - 1))
            for ax in self.grids[grid_index]:
                ax.set_visible(visibility)

    def save(self, annot=None, output_path=None):
        """Saves the collage to disk as an image.

        Parameters
        -----------

        annot : str
            text to annotate the figure with a super title

        output_path : str
            path to save the figure to.
            Note: any spaces in the filename will be replace with ``_``

        """

        if annot is not None:
            self.fig.suptitle(annot, backgroundcolor='black', color='g')

        if output_path is not None:
            output_path = output_path.replace(' ', '_')
            # TODO improve bbox calculations to include ONLY the axes from collage
            # and nothing else
            self.fig.savefig(output_path + '.png', bbox_inches='tight', dpi=200,
                             bbox_extra_artists=self.flat_grid)

    def clear(self):
        """Clears all the axes to start fresh."""

        for ax in self.flat_grid:
            for im_h in ax.findobj(AxesImage):
                im_h.remove()


class MidCollage(Collage):
    """Convenience class to display the mid-slices from all the views."""

    def __init__(self,
                 image,
                 bounding_rect=cfg.bounding_rect_default,
                 fig=None,
                 display_params=None,
                 ):
        """Display mid-slices from all the views.

        image : ndarray
            The image to be attached to the collage, once it is created.
            Must be atleast 3d.

        fig : matplotlib.Figure
            figure handle to create the collage in.
            If not specified, creates a new figure.

        bounding_rect : tuple of 4
            The rectangular area to bind the collage to (in normalized figure coordinates)

        display_params : dict
            dict of keyword parameters that can be passed to matplotlib's `Axes.imshow()`

        """

        super().__init__(view_set=cfg.view_set_default,
                         num_rows=1, num_slices=1, sampler=(50, ),
                         attach_image=image,
                         fig=fig, bounding_rect=bounding_rect,
                         display_params=display_params)


class Carpet(object):
    """Class to unroll the 4D or higher dimensional data into 2D images.

    Typical examples include functional or diffusion MRI data.

    """


    def __init__(self,
                 image_nD,
                 fixed_dim=-1,
                 roi_mask='auto',
                 rescale_data=True,
                 num_frames_to_skip=2,
                 ):
        """
        Constructor

        This class can optionally,


            - can cluster data in the fixed dimension (typically time/gradient dimensions in functional/diffusion MR imaging data)
            - label the resulting clusters.

        Parameters
        ----------

        image_nD : ndarray or str
            input image, or a path to an image, from which the carpet needs to be made.

        fixed_dim : int
            the dimension to be fixed while unrolling the rest. Default: last as indicated by -1

        roi_mask : ndarray or str or None
            if an image of same size as the N-1 dim array, it is interpreted to be a mask to be applied to each 3D/(N-1)D volume
            If its 'auto', an auto background (zeros) mask is computed and removed
            If its None, all the voxels in original image are retained

        rescale_data : bool
            Whether to rescale the input image over the chosen `fixed_dim`
            Default is to rescale to maximize the contrast.

        num_frames_to_skip : int
            number of frames to skip displaying to avoid pixel decimation on screen.
            Choose a number such that to get a total ~600-1000 vertical lines to display.


        """

        self._check_image(image_nD)
        self._add_fixed_dim(fixed_dim)
        self._make_carpet(rescale_data)
        self._apply_mask(roi_mask)

        # choosing mean over median to make it sensitive to outlier values, if any
        self._summary_func = np.mean
        self._carpet_clustered = False

        # TODO option to blur within ROIs, to improve contrast across ROIs?
        #   blurring/additional image processing must be left to classes inheriting from this

        # TODO reorder rows either using anatomical seg, or using clustering

        # dropping alternating voxels if it gets too big
        # to save on memory and avoid losing signal
        if self.carpet.shape[1] > 600 and isinstance(num_frames_to_skip, int):
            print('Too many frames (n={}) to display: '
                  'keeping every {}th frame'
                  ''.format(self.carpet.shape[1], num_frames_to_skip))
            self.carpet = self.carpet[:, ::num_frames_to_skip]


    def _check_image(self, image_nD):
        """Sanity checks on the image data"""

        self.input_image = load_image_from_disk(image_nD)

        if len(self.input_image.shape) < 3:
            raise ValueError('Input image must be atleast 3D')

        if np.count_nonzero(self.input_image) == 0:
            raise ValueError('Input image is completely filled with zeros! '
                             'Must be non-empty')


    def _add_fixed_dim(self, fixed_dim=-1):
        """Makes note of which dimension needs to be fixed, defaulting to last."""

        if fixed_dim in [-1, None, 'last']:
            fixed_dim = len(self.input_image.shape) - 1  # last dimension

        if int(fixed_dim)!=fixed_dim or \
            fixed_dim > len(self.input_image.shape) or \
            fixed_dim < -1:
            raise ValueError('invalid value for the dimension to be fixed!'
                             'Must be an integer in range [0, {}] inclusive'
                             ''.format(len(self.input_image.shape)))

        if self.input_image.shape[fixed_dim] < 2:
            raise ValueError('Input image must have atleast two samples '
                             'in the fixed dimension. It has {}. '
                             'Full image shape: {} '
                             ''.format(self.input_image.shape[fixed_dim],
                                       self.input_image.shape))

        self.fixed_dim = int(fixed_dim)


    def _make_carpet(self, rescale_data):
        """
        Constructs the carpet from the input image.

        Optional rescaling of the data.
        """

        self.carpet = self._unroll_array(self.input_image, self.fixed_dim)
        if rescale_data:
            self.carpet = row_wise_rescale(self.carpet)

    @staticmethod
    def _unroll_array(array, fixed_dim):
        """reshape a given ndarray fixing a chosen dimension (rest gets 'unrolled')."""

        return array.reshape(-1, array.shape[fixed_dim])


    def show(self,
             clustered=False,
             ax_carpet=None,
             label_x_axis='time point',
             label_y_axis='voxels/ROI'):
        """
        Displays the carpet in the given axis.

        Parameters
        ----------

        clustered : bool, optional
            Flag to indicate whether to show the clustered/reduced carpet or the original.
            You must run .cluster_rows_in_roi() before trying to show clustered carpet.

        ax_carpet : Axis, optional
            handle to a valid matplotlib Axis

        label_x_axis : str
            String label for the x-axis of the carpet

        label_y_axis : str
            String label for the y-axis of the carpet

        Returns
        -------

        ax_carpet : Axis
            handle to axis where carpet is shown

        """

        if clustered is True and self._carpet_clustered is False:
            print('You must run .cluster_rows_in_roi() '
                  'before being able to show clustered carpet!')
            return

        if ax_carpet is None:
            self.ax_carpet = plt.gca()
        else:
            if not isinstance(ax_carpet, Axes):
                raise ValueError('Input must be a valid matplotlib Axis!')
            self.ax_carpet = ax_carpet

        plt.sca(self.ax_carpet)
        self.fig = plt.gcf()

        #   vmin/vmax are controlled, because we rescale all to [0, 1]
        self.imshow_params_carpet = dict(interpolation='none', cmap='gray',
                                         aspect='auto', origin='lower', zorder=1)
        # should we control vmin=0.0, vmax=1.0 ??

        if not clustered:
            self.carpet_handle = self.ax_carpet.imshow(self.carpet,
                                                   **self.imshow_params_carpet)
        else:
            self.carpet_handle = self.ax_carpet.imshow(self.clustered_carpet,
                                                       **self.imshow_params_carpet)

        # TODO decorating axes with labels
        self.ax_carpet.set(xlabel=label_x_axis, ylabel=label_y_axis,
                           frame_on=False)
        self.ax_carpet.set_ylim(auto=True)

        return self.ax_carpet


    def save(self, output_path=None, title=None):
        """Saves the current figure with carpet visualization to disk.

        Parameters
        ----------

        output_path : str
            Path to where the figure needs to be saved to.

        title : str
            text to overlay and annotate the visualization (done via plt.suptitle())

        """

        try:
            save_figure(self.fig, output_path=output_path, annot=title)
        except:
            print('Unable to save the figure to disk! \nException: ')
            traceback.print_exc()


    def cluster_rows_in_roi(self,
                            roi_mask=None,
                            num_clusters_per_roi=5,
                            metric='minkowski'):
        """Clusters the data within all the ROIs specified in a mask.

        Parameters
        ----------

        roi_mask : ndarray or None
            volumetric mask defining the list of ROIs, with a label for each voxel.
            This must be the same size in all dimensions except the fixed_dim
            i.e. if you were making a Carpet from an fMRI image of size 125x125x90x400
            fixing the 4th dimension (of size 400), then roi_mask must be of size 125x125x90.

        num_clusters_per_roi : int
            number of clusters (n) to form each ROI specified in the roi_mask
            if n (say 20) is less than number of voxels per a given ROI (say 2000),
            then data from approx. 2000/20=100 voxels would summarized (averaged by default),
            into a single cluster. So if the ROI mask had m ROIs (say 10), then the final clustered carpet
            would have m*n rows (200), regardless of the number of voxels in the 3D image.

        metric : str
            distance metric for the hierarchical clustering algorithm;
            default : 'minkowski'
            Options: anything accepted by `scipy.spatial.distance.pdist`, which can be:
            ‘braycurtis’, ‘canberra’, ‘chebyshev’, ‘cityblock’, ‘correlation’,
            ‘cosine’, ‘dice’, ‘euclidean’, ‘hamming’, ‘jaccard’, ‘kulsinski’,
            ‘mahalanobis’, ‘matching’, ‘minkowski’, ‘rogerstanimoto’, ‘russellrao’,
            ‘seuclidean’, ‘sokalmichener’, ‘sokalsneath’, ‘sqeuclidean’, ‘yule’.


        """

        self._set_roi_mask(roi_mask)

        try:
            clusters = [self._summarize_in_roi(self.roi_mask == label,
                                               num_clusters_per_roi,
                                               metric=metric) for label in self.roi_list]
            self.clustered_carpet = np.vstack(clusters)
        except:
            print('unable to produce the clustered carpet - exception:')
            traceback.print_exc()
            self._carpet_clustered = False
        else:
            self._carpet_clustered = True


    def _set_roi_mask(self, roi_mask):
        """Sets a new ROI mask."""

        if isinstance(roi_mask,
                      np.ndarray):  # not (roi_mask is None or roi_mask=='auto'):
            self._verify_shape_compatibility(roi_mask, 'ROI set')
            self.roi_mask = roi_mask

            self.roi_list = np.unique(roi_mask.flatten())
            np.setdiff1d(self.roi_list, cfg.background_value)
        else:
            self.roi_mask = np.ones(self.carpet.shape[:-1]) # last dim is self.fixed_dim already
            self.roi_list = [1, ] # to allow for iteration


    def _summarize_in_roi(self, label_mask, num_clusters_per_roi=1, metric='minkowski'):
        """returns a single row summarizing (typically via mean) all rows in an ROI."""

        this_label = self.carpet[label_mask.flatten(), :]
        if num_clusters_per_roi == 1:
            out_matrix = self._summary_func(this_label, axis=0)
        else:
            out_matrix = self._make_clusters(this_label, num_clusters_per_roi, metric)

        return out_matrix


    def _make_clusters(self, matrix, num_clusters_per_roi, metric):
        """clusters a given matrix by into specified number of clusters according to given metric"""

        from scipy.cluster.hierarchy import fclusterdata

        # maxclust needed to ensure t is interpreted as # clusters in heirarchical clustering
        group_ids = fclusterdata(matrix, metric=metric, t=num_clusters_per_roi,
                                 criterion='maxclust')
        group_set = np.unique(group_ids)
        clusters = [
            self._summary_func(matrix[group_ids == group, :], axis=0, keepdims=True)
            for group in group_set]

        return np.vstack(clusters).squeeze()


    def _apply_mask(self, roi_mask):
        """Removes voxels outside the given mask or ROI set."""

        # TODO ensure compatible with input image
        #   - must have < N dim and same size in moving dims.
        rows_to_delete = list() # to allow for additional masks to be applied in the future
        if isinstance(roi_mask,
                      np.ndarray):  # not (roi_mask is None or roi_mask=='auto'):
            self._set_roi_mask(roi_mask)

            rows_roi = np.where(self.roi_mask.flatten() == cfg.background_value)

            # TODO below would cause differences in size/shape across mask and carpet!
            self.carpet = np.delete(self.carpet, rows_roi, axis=0)

        else:
            self.roi_mask = np.ones(self.carpet.shape)


    def _verify_shape_compatibility(self, img, img_type):
        """Checks mask shape against input image shape."""

        if self.input_image.shape[:-1] != img.shape:
            raise ValueError('Shape of the {} ({}) is not compatible '
                             'with input image shape: {} '
                             ''.format(img_type, img.shape, self.input_image.shape[:-1]))

    def __str__(self):
        """Helpful repr"""

        return "Carpet {}".format(self.carpet.shape)


    def __repr__(self):

        return "Orig image: {}\n" \
               "Carpet {}".format(self.input_image.shape, self.carpet.shape)


if __name__ == '__main__':
    pass




__all__ = ['get_freesurfer_cmap', ]

import numpy as np
from matplotlib.colors import ListedColormap
import matplotlib.pyplot as plt

def get_freesurfer_cortical_LUT():
    """
    Subset of Freesurfer ColorLUT for cortical labels

    Original at
    https://surfer.nmr.mgh.harvard.edu/fswiki/FsTutorial/AnatomicalROI/FreeSurferColorLUT
    """

    LUT = [ [0, 0, 0],
            [25, 5, 25],
            [25,100,40],
            [125,100,160],
            [100,25,0],
            [120,70,50],
            [220,20,100],
            [220,20,10],
            [180,220,140],
            [220,60,220],
            [180,40,120],
            [140,20,140],
            [20,30,140],
            [35,75,50],
            [225,140,140],
            [200,35,75],
            [160,100,50],
            [20,220,60],
            [60,220,60],
            [220,180,140],
            [20,100,50],
            [220,60,20],
            [120,100,60],
            [220,20,20],
            [220,180,220],
            [60,20,220],
            [160,140,180],
            [80,20,140],
            [75,50,125],
            [20,220,160],
            [20,180,140],
            [140,220,220],
            [80,160,20],
            [100,0,100],
            [70,70,70],
            [150,150,200],
            [220,216,20]
           ]

    LUT = np.array(LUT)

    # to make them range from 0 to 1, required by python (rgb values must be between 0 and 1)
    LUT = LUT/255

    return LUT


def get_freesurfer_subcortical_LUT():
    """
    Subset of Freesurfer ColorLUT for cortical labels

    Original at
    https://surfer.nmr.mgh.harvard.edu/fswiki/FsTutorial/AnatomicalROI/FreeSurferColorLUT
    """

    return plt.cm.Set1 # tab20_r


def get_freesurfer_cmap(vis_type):
    """Provides different colormaps for different visualization types."""

    if vis_type in ('cortical_volumetric', 'cortical_contour'):
            LUT = get_freesurfer_cortical_LUT()
            cmap = ListedColormap(LUT)
    elif vis_type in ('labels_volumetric', 'labels_contour'):
        black = np.array([0, 0, 0, 1])
        cmap = plt.get_cmap('hsv')
        # TODO using more than 20 labels might be a problem?
        cmap = cmap(np.linspace(0, 1, 20))
        # prepending black to paint background as black
        colors = np.vstack((black, cmap))
        cmap = ListedColormap(colors, 'my_colormap')
    else:
        raise NotImplementedError('color map for the visualization type {} has not been implemented!'.format(vis_type))

    return cmap





from genericpath import exists as pexists
from os.path import realpath
from collections import Iterable
import nibabel as nib
import numpy as np
from matplotlib import pyplot as plt
import matplotlib as mpl


def _diff_image(slice1, slice2,
                abs_value=True,
                cmap='gray',
                **kwargs):
    """Computes the difference image"""

    diff = slice1 - slice2

    if abs_value:
        diff = np.abs(diff)

    return diff, cmap


def diff_colormap():
    "Custom colormap to map low values to black or another color."

    # bottom = plt.cm.copper(np.linspace(0., 1, 6))
    black = np.atleast_2d([0., 0., 0., 1.])
    bottom = np.repeat(black, 6, axis=0)
    middle = plt.cm.copper(np.linspace(0, 1, 250))
    # remain = plt.cm.Reds(np.linspace(0, 1, 240))

    colors = np.vstack((bottom, middle))
    diff_colormap = mpl.colors.LinearSegmentedColormap.from_list('diff_colormap', colors)

    return diff_colormap


def get_axis(array, axis, slice_num):
    """Returns a fixed axis"""

    slice_list = [slice(None)] * array.ndim
    slice_list[axis] = slice_num
    slice_data = array[tuple(slice_list)].T  # transpose for proper orientation

    return slice_data


def check_bounding_rect(rect_pos):
    """Ensure the rect spec is valid."""

    if not isinstance(rect_pos, Iterable):
        raise ValueError('rectangle spect must be a tuple of floats '
                         'specifying (left, right, width, height)')

    left, bottom, width, height = rect_pos
    for val, name in zip((left, bottom, width, height),
                         ('left', 'bottom', 'width', 'height')):
        if val < 0.0 or val > 1.0:
            raise ValueError("{}'s value must be >=0 and <= 1.0. "
                             "It is now {}".format(name, val))

    if left + width > 1.0:
        print('rect would extend beyond the width of figure/axis by {}'.format(left + width - 1.0))

    if bottom + height > 1.0:
        print('rect would extend beyond the height of figure/axis by {}'.format(
            bottom + height - 1.0))

    return rect_pos


def check_views(view_set, max_views=3):
    """Ensures valid view/dimensions are selected."""

    if not isinstance(view_set, Iterable):
        view_set = tuple([view_set, ])

    if len(view_set) > max_views:
        raise ValueError('Can only have {} views'.format(max_views))

    return [check_int(view, 'view', min_value=0, max_value=max_views - 1) for view in view_set]


def check_num_slices(num_slices, img_shape=None, num_dims=3):
    """Ensures requested number of slices is valid.

    Atleast 1 and atmost the image size, if available
    """

    if not isinstance(num_slices, Iterable) or len(num_slices) == 1:
        num_slices = np.repeat(num_slices, num_dims)

    if img_shape is not None:
        if len(num_slices) != len(img_shape):
            raise ValueError('The number of dimensions requested is different from image.'
                             ' Must be either 1 or equal to {}'.format(len(img_shape) + 1))
        # upper bounding them to image shape
        num_slices = np.minimum(img_shape, num_slices)

    # lower bounding it to 1
    return np.maximum(1, num_slices)


def check_int(num,
              num_descr='number',
              min_value=0,
              max_value=np.Inf):
    """Validation and typecasting."""

    if not np.isfinite(num) or num < min_value or num > max_value:
        raise ValueError('{}={} is not finite or '
                         'is not >= {} or '
                         'is not < {}'.format(num_descr, num, min_value, max_value))

    return int(num)


def check_patch_size(patch_size):
    """Validation and typcasting"""

    patch_size = np.array(patch_size)
    if patch_size.size == 1:
        patch_size = np.repeat(patch_size, 2).astype('int16')

    return patch_size


def check_params(num_rows, num_cols, padding):
    """Validation and typcasting"""

    num_rows = check_int(num_rows, 'num_rows', min_value=1)
    num_cols = check_int(num_cols, 'num_cols', min_value=1)
    padding = check_int(padding, 'padding', min_value=0)

    return num_rows, num_cols, padding


def read_image(img_spec, bkground_thresh, ensure_num_dim=3):
    """Image reader, with additional checks on size.

    Can optionally remove stray values close to zero (smaller than 5 %ile)."""

    img = load_image_from_disk(img_spec)

    if not np.issubdtype(img.dtype, np.floating):
        img = img.astype('float32')

    if ensure_num_dim == 3:
        img = check_image_is_3d(img)
    elif ensure_num_dim == 4:
        img = check_image_is_4d(img)

    return threshold_image(img, bkground_thresh)


def load_image_from_disk(img_spec):
    """Vanilla image loader."""

    if isinstance(img_spec, str):
        if pexists(realpath(img_spec)):
            hdr = nib.load(img_spec)
            # trying to stick to an orientation
            hdr = nib.as_closest_canonical(hdr)
            img = hdr.get_data()
        else:
            raise IOError('Given path to image does not exist!')
    elif isinstance(img_spec, np.ndarray):
        img = img_spec
    else:
        raise ValueError('Invalid input specified! '
                         'Input either a path to image data, or provide 3d Matrix directly.')

    return img


def check_matching_dims(img_one, img_two):
    """Checks if the dimensions of the two images match, excluding singleton dims."""

    return np.isclose(np.squeeze(img_one.shape), np.squeeze(img_two.shape)).all()


def check_image_is_3d(img):
    """Ensures the image loaded is 3d and nothing else."""

    if len(img.shape) < 3:
        raise ValueError('Input volume must be atleast 3D!')
    elif len(img.shape) == 3:
        for dim_size in img.shape:
            if dim_size < 1:
                raise ValueError('Atleast one slice must exist in each dimension')
    elif len(img.shape) == 4:
        if img.shape[3] != 1:
            raise ValueError('Input volume is 4D with more than one volume!')
        else:
            img = np.squeeze(img, axis=3)
    elif len(img.shape) > 4:
        raise ValueError('Invalid shape of image : {}'.format(img.shape))

    return img


def check_image_is_4d(img, min_num_volumes=2):
    """Ensures the image loaded is 3d and nothing else."""

    if len(img.shape) < 4:
        raise ValueError('Input volume must be 4D!')
    elif len(img.shape) == 4:
        for dim_size in img.shape[:3]:
            if dim_size < 1:
                raise ValueError('Atleast one slice must exist in each dimension')
        if img.shape[3] < min_num_volumes:
            raise ValueError('Input volume is 4D '
                             'with less than {} volumes!'.format(min_num_volumes))
    elif len(img.shape) > 4:
        raise ValueError('Too many dimensions : more than 4.\n'
                         'Invalid shape of image : {}'.format(img.shape))

    return img


def threshold_image(img, bkground_thresh, bkground_value=0.0):
    """
    Thresholds a given image at a value or percentile.

    Replacement value can be specified too.


    Parameters
    -----------
    image_in : ndarray
        Input image

    bkground_thresh : float
        a threshold value to identify the background

    bkground_value : float
        a value to fill the background elements with. Default 0.

    Returns
    -------

    thresholded_image : ndarray
        thresholded and/or filled image

    """

    if bkground_thresh is None:
        return img

    if isinstance(bkground_thresh, str):
        try:
            thresh_perc = float(bkground_thresh.replace('%', ''))
        except:
            raise ValueError(
                'percentile specified could not be parsed correctly '
                ' - must be a string of the form "5%", "10%" etc')
        else:
            thresh_value = np.percentile(img, thresh_perc)
    elif isinstance(bkground_thresh, (float, int)):
        thresh_value = bkground_thresh
    else:
        raise ValueError('Invalid specification for background threshold.')

    img[img < thresh_value] = bkground_value

    return img


def scale_0to1(image_in,
               exclude_outliers_below=False,
               exclude_outliers_above=False):
    """Scale the two images to [0, 1] based on min/max from both.

    Parameters
    -----------
    image_in : ndarray
        Input image

    exclude_outliers_{below,above} : float
        Lower/upper limit, a value between 0 and 100.

    Returns
    -------

    scaled_image : ndarray
        clipped and/or scaled image

    """

    min_value = image_in.min()
    max_value = image_in.max()
    # making a copy to ensure no side-effects
    image = image_in.copy()
    if exclude_outliers_below:
        perctl = float(exclude_outliers_below)
        image[image < np.percentile(image, perctl)] = min_value

    if exclude_outliers_above:
        perctl = float(exclude_outliers_above)
        image[image > np.percentile(image, 100.0 - perctl)] = max_value

    image = (image - min_value) / (max_value - min_value)

    return image


def row_wise_rescale(matrix):
    """
    Row-wise rescale of a given matrix.

    For fMRI data (num_voxels x num_time_points), this would translate to voxel-wise normalization over time.

    Parameters
    ----------

    matrix : ndarray
        Input rectangular matrix, typically a carpet of size num_voxels x num_4th_dim, 4th_dim could be time points or gradients or other appropriate

    Returns
    -------
    normed : ndarray
        normalized matrix

    """

    if matrix.shape[0] <= matrix.shape[1]:
        raise ValueError('Number of voxels is less than the number of time points!! '
                      'Are you sure data is reshaped correctly?')

    min_ = matrix.min(axis=1)
    range_ = matrix.ptp(axis=1)  # ptp : peak to peak, max-min
    min_tile = np.tile(min_, (matrix.shape[1], 1)).T
    range_tile = np.tile(range_, (matrix.shape[1], 1)).T
    # avoiding any numerical difficulties
    range_tile[range_tile < np.finfo(np.float).eps] = 1.0

    normed = (matrix - min_tile) / range_tile

    del min_, range_, min_tile, range_tile

    return normed


def crop_to_seg_extents(img, seg, padding):
    """Crop the image (usually MRI) to fit within the bounding box of a segmentation (or set of seg)"""

    beg_coords, end_coords = crop_coords(seg, padding)

    img = crop_3dimage(img, beg_coords, end_coords)
    seg = crop_3dimage(seg, beg_coords, end_coords)

    return img, seg


def crop_to_extents(img1, img2, padding):
    """Crop the images to ensure both fit within the bounding box"""

    beg_coords1, end_coords1 = crop_coords(img1, padding)
    beg_coords2, end_coords2 = crop_coords(img2, padding)

    beg_coords = np.fmin(beg_coords1, beg_coords2)
    end_coords = np.fmax(end_coords1, end_coords2)

    img1 = crop_3dimage(img1, beg_coords, end_coords)
    img2 = crop_3dimage(img2, beg_coords, end_coords)

    return img1, img2


def crop_image(img, padding=5):
    "Crops an image or slice to its extents"

    if padding < 1:
        return img

    beg_coords, end_coords = crop_coords(img, padding)

    if len(img.shape) == 3:
        img = crop_3dimage(img, beg_coords, end_coords)
    elif len(img.shape) == 2:
        img = crop_2dimage(img, beg_coords, end_coords)
    else:
        raise ValueError('Can only crop 2D or 3D images!')

    return img


def crop_coords(img, padding):
    """Find coordinates describing extent of non-zero portion of image, padded"""

    coords = np.nonzero(img)
    empty_axis_exists = np.any([len(arr) == 0 for arr in coords])
    if empty_axis_exists:
        end_coords = img.shape
        beg_coords = np.zeros((1, img.ndim)).astype(int).flatten()
    else:
        min_coords = np.array([arr.min() for arr in coords])
        max_coords = np.array([arr.max() for arr in coords])
        beg_coords = np.fmax(0, min_coords - padding)
        end_coords = np.fmin(img.shape, max_coords + padding)

    return beg_coords, end_coords


def crop_3dimage(img, beg_coords, end_coords):
    """Crops a 3d image to the bounding box specified."""

    cropped_img = img[
                  beg_coords[0]:end_coords[0],
                  beg_coords[1]:end_coords[1],
                  beg_coords[2]:end_coords[2]
                  ]

    return cropped_img


def crop_2dimage(img, beg_coords, end_coords):
    """Crops a 3d image to the bounding box specified."""

    cropped_img = img[
                  beg_coords[0]:end_coords[0],
                  beg_coords[1]:end_coords[1],
                  ]

    return cropped_img


def pick_slices(img, num_slices_per_view):
    """
    Picks the slices to display in each dimension,
        skipping any empty slices (without any segmentation at all).

    """

    slices = list()
    for view in range(len(img.shape)):
        dim_size = img.shape[view]
        non_empty_slices = np.array(
            [sl for sl in range(dim_size) if np.count_nonzero(get_axis(img, view, sl)) > 0])
        num_non_empty = len(non_empty_slices)

        # trying to 5% slices at the tails (bottom clipping at 0)
        skip_count = max(0, np.around(num_non_empty * 0.05).astype('int16'))
        # only when possible
        if skip_count > 0 and (num_non_empty - 2 * skip_count >= num_slices_per_view):
            non_empty_slices = non_empty_slices[skip_count: -skip_count]
            num_non_empty = len(non_empty_slices)

        # sampling non-empty slices only
        sampled_indices = np.linspace(0, num_non_empty, num=min(num_non_empty, num_slices_per_view),
                                      endpoint=False)
        slices_in_dim = non_empty_slices[np.around(sampled_indices).astype('int64')]

        # ensure you do not overshoot
        slices_in_dim = [sn for sn in slices_in_dim if sn >= 0 or sn <= num_non_empty]

        slices.append(slices_in_dim)

    return slices


def scale_images_0to1(slice1, slice2):
    """Scale the two images to [0, 1] based on min/max from both."""

    min_value = max(slice1.min(), slice2.min())
    max_value = max(slice1.max(), slice2.max())

    slice1 = (slice1 - min_value) / max_value
    slice2 = (slice2 - min_value) / max_value

    return slice1, slice2


def verify_sampler(sampler, image, image_shape, view_set, num_slices):
    """verifies the sampler requested is valid."""

    if isinstance(sampler, str):
        sampler = sampler.lower()
        if sampler not in ['linear', ]:
            raise ValueError('Sampling strategy: {} not implemented.'.format(sampler))
        out_sampler = sampler
        out_sampling_method = 'linear'
    elif isinstance(sampler, Iterable):
        if any([index < 0 or index > 100 for index in sampler]):
            raise ValueError('sampling percentages must be in  [0-100]% range')
        if len(sampler) > min(num_slices):
            num_slices = np.maximum(num_slices, len(sampler))
        out_sampler = np.array(sampler)
        out_sampling_method = 'percentage'
    elif callable(sampler):
        # checking if the callable returns a bool
        for view in view_set:
            middle_slice = int(image_shape[view] / 2)
            if not isinstance(sampler(get_axis(image, view, middle_slice)), bool):
                raise ValueError('sampler callable must return a boolean value (True/False)')

        out_sampler = sampler
        out_sampling_method = 'callable'
    else:
        raise NotImplementedError('Invalid choice for sampler! Choose one of: '
                                  'linear, percentage or callable')

    return out_sampler, out_sampling_method, num_slices


def save_figure(fig, annot=None, output_path=None):

    if annot is not None:
        fig.suptitle(annot, backgroundcolor='black', color='g')

    try:
        fig.tight_layout()
    except:
        pass

    if output_path is not None:
        # output_path = output_path.replace(' ', '_')
        # fig.savefig(output_path + '.png', bbox_inches='tight', dpi=200)
        fig.savefig(output_path, bbox_inches='tight', dpi=200)




view_set_default = (0, 1, 2)
num_slices_default = (10,)
sampler_default = 'linear'

min_density_default = 0.01

num_rows_per_view_default = 2

figsize_default=(14, 10)
bounding_rect_default=(0.02, 0.02, 0.98, 0.98)

num_cols_grid_default = 6
grid_pad_default = 0.01
axis_pad_default = 0.01

background_value = 0



# -*- coding: utf-8 -*-

"""Top-level package for mrivis."""

__author__ = """Pradeep Reddy Raamana"""
__email__ = 'raamana@gmail.com'

__all__ = ['checkerboard', 'color_mix', 'voxelwise_diff', 'collage', 'aseg_on_mri',
           'Collage', 'SlicePicker', 'color_maps']

from sys import version_info
if version_info.major > 2:
    from mrivis.workflow import checkerboard, color_mix, voxelwise_diff, collage, aseg_on_mri
    from mrivis.base import Collage, SlicePicker, MiddleSlicePicker, MidCollage, Carpet
    from mrivis import color_maps
else:
    # from .mrivis import checkerboard
    raise NotImplementedError('mrivis requires Python 3+.')

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



"""
mrivis: Tools to comapre the similarity of two 3d images (structural, functional or parametric maps)

Options include checker board, red green mixer and voxel-wise difference maps.

"""
from functools import partial

from mrivis.base import Collage
from mrivis.color_maps import get_freesurfer_cmap
from mrivis.utils import _diff_image, check_params, check_patch_size, crop_image, \
    crop_to_extents, crop_to_seg_extents, get_axis, pick_slices, read_image, scale_0to1, \
    scale_images_0to1

__all__ = ['checkerboard', 'color_mix', 'voxelwise_diff', 'collage']

import numpy as np
from matplotlib import pyplot as plt, colors, cm
import matplotlib as mpl


def checkerboard(img_spec1=None,
                 img_spec2=None,
                 patch_size=10,
                 view_set=(0, 1, 2),
                 num_slices=(10,),
                 num_rows=2,
                 rescale_method='global',
                 background_threshold=0.05,
                 annot=None,
                 padding=5,
                 output_path=None,
                 figsize=None, ):
    """
    Checkerboard mixer.

    Parameters
    ----------
    img_spec1 : str or nibabel image-like object
        MR image (or path to one) to be visualized

    img_spec2 : str or nibabel image-like object
        MR image (or path to one) to be visualized

    patch_size : int or list or (int, int) or None
        size of checker patch (either square or rectangular)
        If None, number of voxels/patch are chosen such that,
            there will be 7 patches through the width/height.

    view_set : iterable
        Integers specifying the dimensions to be visualized.
        Choices: one or more of (0, 1, 2) for a 3D image

    num_slices : int or iterable of size as view_set
        number of slices to be selected for each view
        Must be of the same length as view_set,
            each element specifying the number of slices for each dimension.
            If only one number is given, same number will be chosen for all dimensions.

    num_rows : int
        number of rows (top to bottom) per each of 3 dimensions

    rescale_method : bool or str or list or None
        Range to rescale the intensity values to
        Default: 'global', min and max values computed based on ranges from both images.
        If false or None, no rescaling is done (does not work yet).

    background_threshold : float or str
        A threshold value below which all the background voxels will be set to zero.
        Default : 0.05. Other option is a string specifying a percentile: '5%', '10%'.
        Specify None if you don't want any thresholding.

    annot : str
        Text to display to annotate the visualization

    padding : int
        number of voxels to pad around each panel.

    output_path : str
        path to save the generate collage to.

    figsize : list
        Size of figure in inches to be passed on to plt.figure() e.g. [12, 12] or [20, 20]

    Returns
    -------
    fig : figure handle
        handle to the collage figure generated.

    """

    img_one, img_two = _preprocess_images(img_spec1,
                                          img_spec2,
                                          rescale_method=rescale_method,
                                          bkground_thresh=background_threshold,
                                          padding=padding)

    display_params = dict(interpolation='none', aspect='auto', origin='lower',
                          cmap='gray', vmin=0.0, vmax=1.0)

    mixer = partial(_checker_mixer, checker_size=patch_size)
    collage = Collage(view_set=view_set, num_slices=num_slices, num_rows=num_rows,
                      figsize=figsize, display_params=display_params)
    collage.transform_and_attach((img_one, img_two), func=mixer)
    collage.save(output_path=output_path, annot=annot)

    return collage


def color_mix(img_spec1=None,
              img_spec2=None,
              alpha_channels=None,
              color_space='rgb',
              view_set=(0, 1, 2),
              num_slices=(10,),
              num_rows=2,
              rescale_method='global',
              background_threshold=0.05,
              annot=None,
              padding=5,
              output_path=None,
              figsize=None, ):
    """
    Color mixer, where each image is represented with a different color (default: red and green) in color channels.

    Parameters
    ----------
    img_spec1 : str or nibabel image-like object
        MR image (or path to one) to be visualized

    img_spec2 : str or nibabel image-like object
        MR image (or path to one) to be visualized

    alpha_channels : (float, float)
        weights for red and green channels in the composite image.
        Default: [1, 1]

    color_space : str
        type of color space to mix the images in.
        Currently supports 'rgb' only.
        Support for 'hsv' coming later.

    view_set : iterable
        Integers specifying the dimensions to be visualized.
        Choices: one or more of (0, 1, 2) for a 3D image

    num_slices : int or iterable of size as view_set
        number of slices to be selected for each view
        Must be of the same length as view_set,
            each element specifying the number of slices for each dimension.
            If only one number is given, same number will be chosen for all dimensions.

    num_rows : int
        number of rows (top to bottom) per each of 3 dimensions

    rescale_method : bool or str or list or None
        Range to rescale the intensity values to
        Default: 'global', min and max values computed based on ranges from both images.
        If false or None, no rescaling is done (does not work yet).

    background_threshold : float or str
        A threshold value below which all the background voxels will be set to zero.
        Default : 0.05. Other option is a string specifying a percentile: '5%', '10%'.
        Specify None if you don't want any thresholding.

    annot : str
        Text to display to annotate the visualization

    padding : int
        number of voxels to pad around each panel.

    output_path : str
        path to save the generate collage to.

    figsize : list
        Size of figure in inches to be passed on to plt.figure() e.g. [12, 12] or [20, 20]

    Returns
    -------
    fig : figure handle
        handle to the collage figure generated.

    """

    if alpha_channels is None:
        alpha_channels = [1, 1]

    if not len(alpha_channels) == 2:
        # not sure if we should make them sum to 1:  or not np.isclose(sum(alpha_channels), 1.0)
        raise ValueError('Alpha must be two elements')

    mixer_params = dict(alpha_channels=alpha_channels,
                        color_space=color_space)
    fig = _compare(img_spec1,
                   img_spec2,
                   num_rows=num_rows,
                   num_cols=num_slices,
                   mixer='color_mix',
                   annot=annot,
                   padding=padding,
                   rescale_method=rescale_method,
                   bkground_thresh=background_threshold,
                   output_path=output_path,
                   figsize=figsize,
                   **mixer_params)

    return fig


def voxelwise_diff(img_spec1=None,
                   img_spec2=None,
                   abs_value=True,
                   cmap='gray',
                   overlay_image=False,
                   overlay_alpha=0.8,
                   num_rows=2,
                   num_cols=6,
                   rescale_method='global',
                   background_threshold=0.05,
                   annot=None,
                   padding=5,
                   output_path=None,
                   figsize=None):
    """
    Voxel-wise difference map.

    Parameters
    ----------
    img_spec1 : str or nibabel image-like object
        MR image (or path to one) to be visualized

    img_spec2 : str or nibabel image-like object
        MR image (or path to one) to be visualized

    abs_value : bool
        Flag indicating whether to take the absolute value of the diffenence or not.
        Default: True, display absolute differences only (so order of images does not matter)

        Colormap to show the difference values.

    overlay_image : bool
        Flag to specify whether to overlay the difference values on the original image.
        .. note: This feature is not reliable and supported well yet.

    num_rows : int
        number of rows (top to bottom) per each of 3 dimensions

    num_cols : int
        number of panels (left to right) per row of each dimension.

    rescale_method : bool or str or list or None
        Range to rescale the intensity values to
        Default: 'global', min and max values computed based on ranges from both images.
        If false or None, no rescaling is done (does not work yet).

    background_threshold : float or str
        A threshold value below which all the background voxels will be set to zero.
        Default : 0.05. Other option is a string specifying a percentile: '5%', '10%'.
        Specify None if you don't want any thresholding.

    annot : str
        Text to display to annotate the visualization

    padding : int
        number of voxels to pad around each panel.

    output_path : str
        path to save the generate collage to.

    figsize : list
        Size of figure in inches to be passed on to plt.figure() e.g. [12, 12] or [20, 20]

    Returns
    -------
    fig : figure handle
        handle to the collage figure generated.


    """

    if not isinstance(abs_value, bool):
        abs_value = bool(abs_value)

    mixer_params = dict(abs_value=abs_value,
                        cmap=cmap,
                        overlay_image=overlay_image,
                        overlay_alpha=overlay_alpha)
    fig = _compare(img_spec1,
                   img_spec2,
                   num_rows=num_rows,
                   num_cols=num_cols,
                   mixer='voxelwise_diff',
                   annot=annot,
                   padding=padding,
                   rescale_method=rescale_method,
                   bkground_thresh=background_threshold,
                   output_path=output_path,
                   figsize=figsize,
                   **mixer_params)

    return fig


def _compare(img_spec1,
             img_spec2,
             num_rows=2,
             num_cols=6,
             mixer='checker_board',
             rescale_method='global',
             annot=None,
             padding=5,
             bkground_thresh=0.05,
             output_path=None,
             figsize=None,
             **kwargs):
    """
    Produces checkerboard comparison plot of two 3D images.

    Parameters
    ----------
    img_spec1 : str or nibabel image-like object
        MR image (or path to one) to be visualized

    img_spec2 : str or nibabel image-like object
        MR image (or path to one) to be visualized

    num_rows : int
        number of rows (top to bottom) per each of 3 dimensions

    num_cols : int
        number of panels (left to right) per row of each dimension.

    mixer : str
        type of mixer to produce the comparison figure.
        Options: checker_board, color_mix, diff_abs,

    rescale_method : bool or str or list or None
        Method to rescale the intensity values to.
        Choices : 'global', 'each', False or None.

        Default: 'global', min and max values computed based on ranges from both images.
        If 'each', rescales each image separately to [0, 1].
            This option is useful when overlaying images with very different intensity ranges e.g. from different modalities altogether.
        If False or None, no rescaling is done (does not work yet).

    annot : str
        Text to display to annotate the visualization

    padding : int
        number of voxels to pad around each panel.

    output_path : str
        path to save the generate collage to.

    figsize : list
        Size of figure in inches to be passed on to plt.figure() e.g. [12, 12] or [20, 20]

    kwargs : dict
        Additional arguments specific to the particular mixer
        e.g. alpha_channels = [1, 1] for the color_mix mixer

    Returns
    -------

    """

    num_rows, num_cols, padding = check_params(num_rows, num_cols, padding)

    img1, img2 = check_images(img_spec1, img_spec2, bkground_thresh=bkground_thresh)
    img1, img2 = crop_to_extents(img1, img2, padding)

    num_slices_per_view = num_rows * num_cols
    slices = pick_slices(img2, num_slices_per_view)

    rescale_images, img1, img2, min_value, max_value = check_rescaling(img1, img2, rescale_method)

    plt.style.use('dark_background')

    num_axes = 3
    if figsize is None:
        figsize = [3 * num_axes * num_rows, 3 * num_cols]
    fig, ax = plt.subplots(num_axes * num_rows, num_cols, figsize=figsize)

    # displaying some annotation text if provided
    # good choice would be the location of the input images (for future refwhen image is shared or misplaced!)
    if annot is not None:
        fig.suptitle(annot, backgroundcolor='black', color='g')

    display_params = dict(interpolation='none', aspect='equal', origin='lower')

    ax = ax.flatten()
    ax_counter = 0
    for dim_index in range(3):
        for slice_num in slices[dim_index]:
            plt.sca(ax[ax_counter])
            ax_counter = ax_counter + 1

            slice1 = get_axis(img1, dim_index, slice_num)
            slice2 = get_axis(img2, dim_index, slice_num)

            mixed, mixer_spec_params = _generic_mixer(slice1, slice2, mixer, **kwargs)
            display_params.update(mixer_spec_params)

            plt.imshow(mixed, vmin=min_value, vmax=max_value, **display_params)

            # adjustments for proper presentation
            plt.axis('off')

    fig.tight_layout()

    if output_path is not None:
        output_path = output_path.replace(' ', '_')
        fig.savefig(output_path + '.png', bbox_inches='tight')

    # plt.close()

    return fig


def _preprocess_images(img_spec1,
                       img_spec2,
                       rescale_method='global',
                       bkground_thresh=None,
                       padding=5,
                       ):

    img_one, img_two = check_images(img_spec1, img_spec2, bkground_thresh=bkground_thresh)
    img_one, img_two = crop_to_extents(img_one, img_two, padding)

    rescale_images, img_one, img_two, \
    min_value, max_value = check_rescaling(img_one, img_two, rescale_method)

    return img_one, img_two


def _open_figure(slicer, num_rows_per_view, figsize=(15, 11)):

    total_num_rows = len(slicer.view_set) * num_rows_per_view
    total_num_panels = sum(slicer.num_slices)
    num_cols = int(np.ceil(total_num_panels / total_num_rows))

    plt.style.use('dark_background')
    fig, axes = plt.subplots(total_num_rows, num_cols, figsize=figsize,
                             subplot_kw=dict(rasterized=True),
                             gridspec_kw=dict(wspace=0.005, hspace=0.005))
    axes = axes.flatten()
    # plt.subplots_adjust(wspace=0.005, hspace=0.005, top=0.01, bottom=0.0)

    for ix, ax in enumerate(axes):
        ax.axis('off')

    return fig, axes


def collage(img_spec,
            num_rows=2,
            num_cols=6,
            rescale_method='global',
            cmap='gray',
            annot=None,
            padding=5,
            bkground_thresh=None,
            output_path=None,
            figsize=None,
            **kwargs):
    "Produces a collage of various slices from different orientations in the given 3D image"

    num_rows, num_cols, padding = check_params(num_rows, num_cols, padding)

    img = read_image(img_spec, bkground_thresh=bkground_thresh)
    img = crop_image(img, padding)

    img, (min_value, max_value) = check_rescaling_collage(img, rescale_method,
                                                          return_extrema=True)
    num_slices_per_view = num_rows * num_cols
    slices = pick_slices(img, num_slices_per_view)

    plt.style.use('dark_background')

    num_axes = 3
    if figsize is None:
        figsize = [3 * num_axes * num_rows, 3 * num_cols]
    fig, ax = plt.subplots(num_axes * num_rows, num_cols, figsize=figsize)

    # displaying some annotation text if provided
    if annot is not None:
        fig.suptitle(annot, backgroundcolor='black', color='g')

    display_params = dict(interpolation='none', cmap=cmap,
                          aspect='equal', origin='lower',
                          vmin=min_value, vmax=max_value)

    ax = ax.flatten()
    ax_counter = 0
    for dim_index in range(3):
        for slice_num in slices[dim_index]:
            plt.sca(ax[ax_counter])
            ax_counter = ax_counter + 1
            slice1 = get_axis(img, dim_index, slice_num)
            # slice1 = crop_image(slice1, padding)
            plt.imshow(slice1, **display_params)
            plt.axis('off')

    fig.tight_layout()

    if output_path is not None:
        output_path = output_path.replace(' ', '_')
        fig.savefig(output_path + '.png', bbox_inches='tight')

    # plt.close()

    return fig


def aseg_on_mri(mri_spec,
                aseg_spec,
                alpha_mri=1.0,
                alpha_seg=1.0,
                num_rows=2,
                num_cols=6,
                rescale_method='global',
                aseg_cmap='freesurfer',
                sub_cortical=False,
                annot=None,
                padding=5,
                bkground_thresh=0.05,
                output_path=None,
                figsize=None,
                **kwargs):
    "Produces a collage of various slices from different orientations in the given 3D image"

    num_rows, num_cols, padding = check_params(num_rows, num_cols, padding)

    mri = read_image(mri_spec, bkground_thresh=bkground_thresh)
    seg = read_image(aseg_spec, bkground_thresh=0)
    mri, seg = crop_to_seg_extents(mri, seg, padding)

    num_slices_per_view = num_rows * num_cols
    slices = pick_slices(seg, num_slices_per_view)

    plt.style.use('dark_background')

    num_axes = 3
    if figsize is None:
        figsize = [5 * num_axes * num_rows, 5 * num_cols]
    fig, ax = plt.subplots(num_axes * num_rows, num_cols, figsize=figsize)

    # displaying some annotation text if provided
    if annot is not None:
        fig.suptitle(annot, backgroundcolor='black', color='g')

    display_params_mri = dict(interpolation='none', aspect='equal', origin='lower',
                              cmap='gray', alpha=alpha_mri,
                              vmin=mri.min(), vmax=mri.max())
    display_params_seg = dict(interpolation='none', aspect='equal', origin='lower',
                              alpha=alpha_seg)

    normalize_labels = colors.Normalize(vmin=seg.min(), vmax=seg.max(), clip=True)
    fs_cmap = get_freesurfer_cmap(sub_cortical)
    label_mapper = cm.ScalarMappable(norm=normalize_labels, cmap=fs_cmap)

    ax = ax.flatten()
    ax_counter = 0
    for dim_index in range(3):
        for slice_num in slices[dim_index]:
            plt.sca(ax[ax_counter])
            ax_counter = ax_counter + 1

            slice_mri = get_axis(mri, dim_index, slice_num)
            slice_seg = get_axis(seg, dim_index, slice_num)

            # # masking data to set no-value pixels to transparent
            # seg_background = np.isclose(slice_seg, 0.0)
            # slice_seg = np.ma.masked_where(seg_background, slice_seg)
            # slice_mri = np.ma.masked_where(np.logical_not(seg_background), slice_mri)

            seg_rgb = label_mapper.to_rgba(slice_seg)
            plt.imshow(seg_rgb, **display_params_seg)
            plt.imshow(slice_mri, **display_params_mri)
            plt.axis('off')

    # plt.subplots_adjust(wspace=0.0, hspace=0.0)
    plt.subplots_adjust(left=0.01, right=0.99,
                        bottom=0.01, top=0.99,
                        wspace=0.05, hspace=0.02)
    # fig.tight_layout()

    if output_path is not None:
        output_path = output_path.replace(' ', '_')
        fig.savefig(output_path + '.png', bbox_inches='tight')

    # plt.close()

    return fig


def _generic_mixer(slice1, slice2, mixer_name, **kwargs):
    """
    Generic mixer to process two slices with appropriate mixer
        and return the composite to be displayed.
    """

    mixer_name = mixer_name.lower()
    if mixer_name in ['color_mix', 'rgb']:
        mixed = _mix_color(slice1, slice2, **kwargs)
        cmap = None  # data is already RGB-ed
    elif mixer_name in ['checkerboard', 'checker', 'cb', 'checker_board']:
        checkers = _get_checkers(slice1.shape, **kwargs)
        mixed = _checker_mixer(slice1, slice2, checkers)
        cmap = 'gray'
    elif mixer_name in ['diff', 'voxelwise_diff', 'vdiff']:

        mixed, cmap = _diff_image(slice1, slice2, **kwargs)
        # if kwargs['overlay_image'] is True:
        #     diff_cmap = diff_colormap()
        #     plt.imshow(slice1, alpha=kwargs['overlay_alpha'], **display_params)
        #     plt.hold(True)
        #     plt.imshow(mixed,
        #                cmap=diff_cmap,
        #                vmin=min_value, vmax=max_value,
        #                **display_params)
        # else:
        #     plt.imshow(mixed, cmap=cmap,
        #                vmin=min_value, vmax=max_value,
        #                **display_params)

    else:
        raise ValueError('Invalid mixer name chosen.')

    disp_params = dict(cmap=cmap)

    return mixed, disp_params


def check_rescaling_collage(img, rescale_method=None,
                            return_extrema=True):
    ""

    if not (isinstance(rescale_method, str) or rescale_method is None):
        raise ValueError('rescale_method method be "global", "slice" or None')

    if rescale_method is None:
        return img, rescale_method

    rescale_method = rescale_method.lower()
    if rescale_method in ['global']:
        img = scale_0to1(img)

    if return_extrema:
        min_value = img.min()
        max_value = img.max()
        return img, (min_value, max_value)
    else:
        return img


def check_rescaling(img1, img2, rescale_method):
    """Estimates the intensity range to clip the visualizations to"""

    # estimating intensity ranges
    if rescale_method is None:
        # this section is to help user to avoid all intensity rescaling altogther!
        # TODO bug does not work yet, as pyplot does not offer any easy way to control it
        rescale_images = False
        min_value = None
        max_value = None
        norm_image = None  # mpl.colors.NoNorm doesn't work yet. data is getting linearly normalized to [0, 1]
    elif isinstance(rescale_method, str):
        if rescale_method.lower() in ['global']:
            # TODO need a way to alert the user if one of the distributions is too narrow
            #  in which case that image will be collapsed to an uniform value
            combined_distr = np.concatenate((img1.flatten(), img2.flatten()))
            min_value = combined_distr.min()
            max_value = combined_distr.max()
        elif rescale_method.lower() in ['each']:
            img1 = scale_0to1(img1)
            img2 = scale_0to1(img2)
            min_value = 0.0
            max_value = 1.0
        else:
            raise ValueError('rescaling method can only be "global" or "each"')

        rescale_images = True
        norm_image = mpl.colors.Normalize
    elif len(rescale_method) == 2:
        min_value = min(rescale_method)
        max_value = max(rescale_method)
        rescale_images = True
        norm_image = mpl.colors.Normalize
    else:
        raise ValueError('Invalid intensity range!. It must be either : '
                         '1) a list/tuple of two distinct values or'
                         '2) "global" indicating rescaling based on min/max values derived from both images or'
                         '3) None, no rescaling or norming altogether. ')

    return rescale_images, img1, img2, min_value, max_value


def check_images(img_spec1, img_spec2, bkground_thresh=0.05):
    """Reads the two images and assers identical shape."""

    img1 = read_image(img_spec1, bkground_thresh)
    img2 = read_image(img_spec2, bkground_thresh)

    if img1.shape != img2.shape:
        raise ValueError('size mismatch! First image: {} Second image: {}\n'
                         'Two images to be compared must be of the same size in all dimensions.'.format(
            img1.shape,
            img2.shape))

    return img1, img2


def _get_checkers(slice_shape, patch_size):
    """Creates checkerboard of a given tile size, filling a given slice."""

    if patch_size is not None:
        patch_size = check_patch_size(patch_size)
    else:
        # 7 patches in each axis, min voxels/patch = 3
        # TODO make 7 a user settable parameter
        patch_size = np.round(np.array(slice_shape) / 7).astype('int16')
        patch_size = np.maximum(patch_size, np.array([3, 3]))

    black = np.zeros(patch_size)
    white = np.ones(patch_size)
    tile = np.vstack((np.hstack([black, white]), np.hstack([white, black])))

    # using ceil so we can clip the extra portions
    num_tiles = np.ceil(np.divide(slice_shape, tile.shape)).astype(int)
    checkers = np.tile(tile, num_tiles)

    # clipping any extra columns or rows
    if any(np.greater(checkers.shape, slice_shape)):
        if checkers.shape[0] > slice_shape[0]:
            checkers = np.delete(checkers, np.s_[slice_shape[0]:], axis=0)
        if checkers.shape[1] > slice_shape[1]:
            checkers = np.delete(checkers, np.s_[slice_shape[1]:], axis=1)

    return checkers


def _mix_color(slice1, slice2, alpha_channels, color_space):
    """Mixing them as red and green channels"""

    if slice1.shape != slice2.shape:
        raise ValueError('size mismatch between cropped slices and checkers!!!')

    alpha_channels = np.array(alpha_channels)
    if len(alpha_channels) != 2:
        raise ValueError('Alphas must be two value tuples.')

    slice1, slice2 = scale_images_0to1(slice1, slice2)

    # masking background
    combined_distr = np.concatenate((slice1.flatten(), slice2.flatten()))
    image_eps = np.percentile(combined_distr, 5)
    background = np.logical_or(slice1 <= image_eps, slice2 <= image_eps)

    if color_space.lower() in ['rgb']:

        red = alpha_channels[0] * slice1
        grn = alpha_channels[1] * slice2
        blu = np.zeros_like(slice1)

        # foreground = np.logical_not(background)
        # blu[foreground] = 1.0

        mixed = np.stack((red, grn, blu), axis=2)

    elif color_space.lower() in ['hsv']:

        raise NotImplementedError(
            'This method (color_space="hsv") is yet to fully conceptualized and implemented.')

        # TODO other ideas: hue/saturation/intensity value driven by difference in intensity?
        hue = alpha_channels[0] * slice1
        sat = alpha_channels[1] * slice2
        val = np.ones_like(slice1)

        hue[background] = 1.0
        sat[background] = 0.0
        val[background] = 0.0

        mixed = np.stack((hue, sat, val), axis=2)
        # converting to RGB
        mixed = mpl.colors.hsv_to_rgb(mixed)

    # ensuring all values are clipped to [0, 1]
    mixed[mixed <= 0.0] = 0.0
    mixed[mixed >= 1.0] = 1.0

    return mixed


def _checker_mixer(slice1,
                   slice2,
                   checker_size=None):
    """Mixes the two slices in alternating areas specified by checkers"""

    checkers = _get_checkers(slice1.shape, checker_size)
    if slice1.shape != slice2.shape or slice2.shape != checkers.shape:
        raise ValueError('size mismatch between cropped slices and checkers!!!')

    mixed = slice1.copy()
    mixed[checkers > 0] = slice2[checkers > 0]

    return mixed


def cli_run():
    raise NotImplementedError


if __name__ == '__main__':
    cli_run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
import versioneer

with open('README.rst') as readme_file:
    readme = readme_file.read()


requirements = [
    'nibabel',
    'matplotlib',
    'numpy',
    'scipy',
]

setup_requirements = [
    'pytest-runner',
    # TODO(raamana): put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    'pytest',
    # TODO: put package test requirements here
]

setup(
    name='mrivis',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Tools and scripts for visualization and comparison of 3d MRI scans (T1, T2 etc)",
    long_description=readme,
    author="Pradeep Reddy Raamana",
    author_email='raamana@gmail.com',
    url='https://github.com/raamana/mrivis',
    packages=find_packages(include=['mrivis']),
    entry_points={
        'console_scripts': [
            'mrivis=mrivis.cli:main'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='mrivis',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)
